# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from packaging.version import Version

from smriprep import (
    __copyright__ as _copyright,
)
from smriprep import (
    __package__ as _package,
)
from smriprep import (
    __version__ as _version,
)

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), 'sphinxext')))

from github_link import make_linkcode_resolve  # noqa: E402

os.environ['NO_ET'] = '1'

sys.path.insert(0, os.path.abspath('../wrapper'))

# -- Project information -----------------------------------------------------
project = _package
copyright = _copyright  # noqa: A001
author = 'The sMRIPrep Developers'

# The short X.Y version
version = Version(_version).public
# The full version, including alpha/beta/rc tags
release = version


# -- General configuration ---------------------------------------------------
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinxarg.ext',  # argparse extension
    'sphinxcontrib.apidoc',
    'nipype.sphinxext.plot_workflow',
    'nipype.sphinxext.apidoc',
]

autodoc_mock_imports = [
    'matplotlib',
    'nilearn',
    'nitime',
    'pandas',
    'seaborn',
    'skimage',
    'svgutils',
    'transforms3d',
]

# Accept custom section names to be parsed for numpy-style docstrings
# of parameters.
# Requires pinning sphinxcontrib-napoleon to a specific commit while
# https://github.com/sphinx-contrib/napoleon/pull/10 is merged.
napoleon_use_param = False
napoleon_custom_sections = [
    ('Inputs', 'Parameters'),
    ('Outputs', 'Parameters'),
    ('Attributes', 'Parameters'),
    ('Mandatory Inputs', 'Parameters'),
    ('Optional Inputs', 'Parameters'),
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
    'api/modules.rst',
    'api/smriprep.rst',
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'smriprepdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        'smriprep.tex',
        'sMRIPrep Documentation',
        'The NiPreps Developers',
        'manual',
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'smriprep', 'sMRIPrep Documentation', [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        'smriprep',
        'sMRIPrep Documentation',
        author,
        'sMRIPrep',
        'One line description of project.',
        'Miscellaneous',
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

apidoc_module_dir = '../smriprep'
apidoc_output_dir = 'api'
apidoc_excluded_paths = [
    'conftest.py',
    '*/conftest.py',
    '*/tests/*',
    'tests/*',
    'data/*',
    'conf/*',
]
apidoc_separate_modules = True
apidoc_extra_args = ['--module-first', '-d 1', '-T']

# Options for github links
# The following is used by sphinx.ext.linkcode to provide links to github
linkcode_resolve = make_linkcode_resolve(
    'smriprep',
    'https://github.com/nipreps/smriprep/blob/{revision}/{package}/{path}#L{lineno}',
)

# -- Options for intersphinx extension ---------------------------------------
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    'matplotlib': ('https://matplotlib.org/', None),
    'bids': ('https://bids-standard.github.io/pybids/', None),
    'nibabel': ('https://nipy.org/nibabel/', None),
    'nipype': ('https://nipype.readthedocs.io/en/latest/', None),
    'niworkflows': ('https://www.nipreps.org/niworkflows/', None),
    'templateflow': ('https://www.templateflow.org/python-client', None),
}

# -- Options for versioning extension ----------------------------------------
scv_show_banner = True
